/*
 * cfgtool - A Configurator tool for ROSAND Technologies products.
 *
 * Inspired from AVRDUDE program.
 *
 * Copyright (C) 2000-2005  Brian S. Dean <bsd@bsdhome.com>
 * Copyright 2007-2009 Joerg Wunsch <j@uriah.heep.sax.de>
 * Copyright (c) 2011, Rosand Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#include "config.h"
#include "update.h"

UPDATE * parse_op(char * s)
{
  char buf[1024];
  char * p, * cp, c, * last;
  UPDATE * upd;
  int i;
  size_t fnlen;
  int len;

  upd = (UPDATE *)malloc(sizeof(UPDATE));
  if (upd == NULL) {
    fprintf(stderr, "%s: out of memory\n", progname);
    exit(1);
  }

  i = 0;
  p = s;
  len = strlen(p);
  last = p+len-1;
  if(*last != ':') {
	fprintf(stderr, "%s: invalid argument in update operation\n",
			progname, *last);
	fprintf(stderr,	"%sdelimiter \":\" is missing.\n",
			progbuf);
	free(upd);
	return NULL;
  }
  while ((i < (sizeof(buf)-1) && *p && (*p != ':')))
    buf[i++] = *p++;
  buf[i] = 0;

  upd->memtype = (char *)malloc(strlen(buf)+1);
  if (upd->memtype == NULL) {
    fprintf(stderr, "%s: out of memory\n", progname);
    exit(1);
  }
  strcpy(upd->memtype, buf);

  p++;
  if (*p == 'r') {
    upd->op = DEVICE_READ;
  }
  else if (*p == 'w') {
    upd->op = DEVICE_WRITE;
  }
  else if (*p == 'a') {
	  upd->op = DEVICE_ADD;
  }
  else if (*p == 'd') {
	  upd->op = DEVICE_DELETE;
  }
  else {
    fprintf(stderr, "%s: invalid I/O mode '%c' in update operation\n",
            progname, *p);
    fprintf(stderr,
            "%sallowed values are:\n"
            "%s    r = read configuration\n"
            "%s    w = write configuration\n"
            "%s    a = add configuration\n"
            "%s    d = delete configuration\n",
            progbuf, progbuf, progbuf, progbuf, progbuf, progbuf);
    free(upd->memtype);
    free(upd);
    return NULL;
  }

  p++;

  if (*p != ':') {
    fprintf(stderr, "%s: invalid operation\n", progname);
    free(upd->memtype);
    free(upd);
    return NULL;
  }

  p++;

  /*
   * Now, parse the prefix index and value components.
   */

  switch(upd->op) {
  case DEVICE_READ:
  case DEVICE_DELETE:
	cp = p;
	if(cp > last) {
	  upd->format = 0;
	  fnlen = 0;
	  upd->filename = (char *)malloc(fnlen + 1);
	}
	else {
	  p = strchr(cp, ':');
	  if(p == last) {
		upd->format = atoi(cp);
		fnlen = 0;
		upd->filename = (char *)malloc(fnlen + 1);
		if (upd->filename == NULL) {
			fprintf(stderr, "%s: out of memory\n", progname);
			exit(1);
		}
	  }
	  else {
		fprintf(stderr, "%s: invalid prefix index '%s' in update specifier\n",
				progname, p);
		free(upd->memtype);
		free(upd);
		return NULL;
	  }
	}
	break;
  case DEVICE_WRITE:
	if(*p == 0) {
	  fprintf(stderr, "%s: invalid argument in update operation\n",
			  progname);
	  free(upd->memtype);
	  free(upd);
	  return NULL;
	}
	fnlen = strlen(p)-1;
	upd->filename = (char *)malloc(fnlen + 1);
	if (upd->filename == NULL) {
		fprintf(stderr, "%s: out of memory\n", progname);
		exit(1);
	}
	memcpy(upd->filename, p, fnlen);
	break;
  case DEVICE_ADD:
	if(*p == 0) {
	  fprintf(stderr, "%s: invalid argument in update operation\n",
			  progname);
	  free(upd->memtype);
	  free(upd);
	  return NULL;
	}
	cp = p;
	p = strchr(cp, ':');
	*p = 0;
	upd->format = atoi(cp);
	p++;
	fnlen = strlen(p)-1;
	upd->filename = (char *)malloc(fnlen + 1);
	if (upd->filename == NULL) {
		fprintf(stderr, "%s: out of memory\n", progname);
		exit(1);
	}
	memcpy(upd->filename, p, fnlen);
	break;
  }
  upd->filename[fnlen] = 0;

  return upd;
}

UPDATE * dup_update(UPDATE * upd)
{
  UPDATE * u;

  u = (UPDATE *)malloc(sizeof(UPDATE));
  if (u == NULL) {
    fprintf(stderr, "%s: out of memory\n", progname);
    exit(1);
  }

  memcpy(u, upd, sizeof(UPDATE));

  u->memtype = strdup(upd->memtype);
  u->filename = strdup(upd->filename);

  return u;
}

int do_op(PROGRAMMER * pgm, struct part * p, UPDATE * upd, int nowrite,
          int verify)
{
  PART * v;
  MEM * mem;
  int size;
  int rc;
  int do_check;

  mem = locate_mem(p, upd->memtype);
  if (mem == NULL) {
    fprintf(stderr, "\"%s\" parameter not defined for product \"%s\"\n",
            upd->memtype, p->desc);
    return -1;
  }

  if (upd->op == DEVICE_READ){
	unsigned char *buf;
	unsigned int n;
	int i;

    /*
     * read out the specified device configuration
     */
    rc = part_read(pgm, p, upd->memtype, 0, upd->format);
    if (rc < 0) {
      fprintf(stderr, "%s: failed to read %s, rc=%d\n",
              progname, mem->desc, rc);
      return -1;
    }
    size = rc;
    fprintf(stderr, "%s%s ",
          progbuf, mem->desc);
    if(upd->format) fprintf(stderr, "%d ",
          upd->format);
    fprintf(stderr, "= ");

    switch(mem->mode) {
    case 0:
    	buf = mem->buf;
    	n = i = 0;
    	do {
    		n |= (*buf << (8*i));
    		i++;
    		buf++;
    	} while(buf - mem->buf < size);
    	fprintf(stderr, "%d\n", n);
    	break;
    case 1:
    	buf = mem->buf;
    	n = i = 0;
    	do {
    		n |= (*buf << (8*i));
    		i++;
    		buf++;
    	} while(buf - mem->buf < size);
    	fprintf(stderr, "0x%x\n", n);
    	break;
    case 2:
    	buf = mem->buf;
		n = i = 0;
    	do {
    		n |= (*buf << (8*i));
    		i++;
    		buf++;
    		if(i % 2 == 0) {
    			fprintf(stderr, "%x:", n);
    			n = i = 0;
    		}
    	} while(buf - mem->buf < size);
    	fprintf(stderr, "\n");
    	break;
    case 3:
    	fprintf(stderr, "%s", mem->buf);
    	break;
    default:
    	fprintf(stderr, "invalid display format %d\n",
    			mem->mode);
    	break;
    }
  }
  else if ((upd->op == DEVICE_WRITE) || (upd->op == DEVICE_ADD)) {
	unsigned char * pt;
	double   n;
	int    i, val;
	unsigned char buf[10];
	char *cp;
	char *c;
	int   opt;

	cp = upd->filename;
	mem->buf = (unsigned char *)malloc(256);
	memset(mem->buf, 0, 256);
	switch(mem->mode) {
	case 0:
	case 1:
		n = strtod(cp, &c);
		if(*c != 0) {   /* no conversion performed */
		  char temp;
		  if (upd->op == DEVICE_WRITE) {
		    temp = 'w';
		  } else {
		    temp = 'a';
		  }
		  fprintf(stderr, "%s: invalid argument '%s:%c:%s'\n",
				  progname, upd->memtype,temp, upd->filename);
		  return -1;
		}
		pt = mem->buf;
		val = n;
		for(i = 0; i < mem->size; i++) {
			*pt++ = (val >> (8*i)) & 0xff;
		}
		opt = 0;
		break;
	case 2:
		pt = mem->buf;
		{
			unsigned char *tmp = buf;
			char *p;

			*tmp++ = '0';
			*tmp++ = 'x';
			do {
				p = strchr(cp, ':');
				if(p == NULL) {
					strcpy(tmp, cp);
				} else {
					strncpy(tmp, cp, p-cp);
					*(tmp+(p-cp)) = 0;
					cp = p + 1;
					tmp = buf + 2;
				}
				n = strtod(buf, &c);
				if(*c != 0) {   /* no conversion performed */
		                  char temp;
		                  if (upd->op == DEVICE_WRITE) {
		                    temp = 'w';
		                  } else {
		                    temp = 'a';
		                  }
				  fprintf(stderr, "%s: invalid argument '%s:%c:%s'\n",
						  progname, upd->memtype,temp, upd->filename);
				  return -1;
				}
				val = n;
				for(i = 0; i < 2; i++) {
					*pt++ = (val >> (8*i)) & 0xff;
				}
			} while(p != NULL);
		}
		opt = 0;
		break;
	default:
		strcpy(mem->buf, cp);
		opt = strlen(cp);
		break;
	}

    /* write the selected device configuration */
	if(nowrite) {
		rc = mem->mode > 2 ? opt:mem->size;
	}
	else {
	    rc = part_write(pgm, p, upd->memtype, opt, upd->format);
	}
    if (rc < 0) {
      fprintf(stderr, "%s: failed to write %s, rc=%d\n",
              progname, mem->desc, rc);
      return -1;
    }

    if (quell_progress < 2) {
      fprintf(stderr, "%s: %d bytes of %s written\n", progname,
            rc, mem->desc);
    }
  }
  else if (upd->op == DEVICE_VERIFY) {
    /*
     * verify that the in memory file (p->mem[M_FLASH])
     * is the same as what is on the chip
     */
    pgm->vfy_led(pgm, ON);

    v = dup_part(p);

    if (quell_progress < 2) {
      fprintf(stderr, "%s: verifying %s memory:\n",
            progname, mem->desc);
    }

    mem = locate_mem(p, upd->memtype);
    if(mem->mode == 3) {
        rc = strlen(mem->buf);
    } else {
        rc = mem->size;
    }
    size = rc;
    if (quell_progress < 2) {
      fprintf(stderr, "%s: reading on-chip %s data:\n",
            progname, mem->desc);
    }

    rc = part_read(pgm, v, upd->memtype, 0, upd->format);
    if (rc < 0) {
      fprintf(stderr, "%s: failed to read all of %s memory, rc=%d\n",
              progname, mem->desc, rc);
      pgm->err_led(pgm, ON);
      return -1;
    }

    if (quell_progress < 2) {
      fprintf(stderr, "%s: verifying ...\n", progname);
    }
    rc = part_verify(p, v, upd->memtype, size);
    if (rc < 0) {
      fprintf(stderr, "%s: verification error; content mismatch\n",
              progname);
      pgm->err_led(pgm, ON);
      return -1;
    }

    if (quell_progress < 2) {
      fprintf(stderr, "%s: %d bytes of %s verified\n",
              progname, rc, mem->desc);
    }

    pgm->vfy_led(pgm, OFF);
  }
  else if (upd->op == DEVICE_DELETE) {

#define CMD_DELETE       0x05

	unsigned char command[4];
	unsigned char res[4];
	unsigned char *cp = command + 1;
	int rc;

	*cp++ = CMD_DELETE;
	*cp++ = mem->pollindex;
	if(upd->format) *cp++ = upd->format;
	command[0] = (cp - command) - 1;
	rc = pgm->cmd(pgm, command, res);
	if(rc < 0) {
      fprintf(stderr, "%s: failed to delete %s memory, rc=%d\n",
    		  progname, mem->desc, rc);
	}
  }
  else {
    fprintf(stderr, "%s: invalid update operation (%d) requested\n",
            progname, upd->op);
    return -1;
  }

  return 0;
}

